alignment_help = '''
BSBolt Align -F1 {fastq1} -DB {BSBolt db} -O {output}

-h, --help    show this help message and exit

Input / Output Options:
  -F1 File     path to fastq 1
  -F2 File     path to fastq 2 [null]
  -O File      output Prefix
  -DB File     path to BSBolt database
  -R Str       read group header line such as '@RG ID:foo SM:bar' [null]
  -H Str       insert STR to header if it starts with @; or insert lines in FILE [null]
  -XA Int,Int  if there are <INT hits with score >80 percent of the max score, output all in XA [100,200]
  -DR Float    drop ratio for alternative hits reported in XA tag, 
               for best bisulfite alignment performance set at or above default [0.95]
  -p           smart pairing (ignoring in2.fq)
Scoring Options
  -A Int       score for a sequence match, which scales options -TdBOELU unless overridden [1]
  -B Int       penalty for a mismatch [4]
  -INDEL Int   gap open penalties for deletions and insertions [6,6]
  -E Int       gap extension penalty; a gap of size k cost '{-O} + {-E}*k' [1,1]
  -L Int,Int   penalty for 5'- and 3'-end clipping [30,30]
  -U Int       penalty for an unpaired read pair [17]
Bisulfite Options
  -UN          library undirectional, ie. consider PCR products of bisulfite converted DNA
  -CP Float    CH conversion proportion threshold [0.5]
  -CT Int      number of CH sites needed to assess read conversion
  -SP Float    substitution threshold for read bisulfite conversion patterns (ie C2T, G2A) [0.1]
               for undirectional libraries the substitution pattern with the fewer number of 
               substitutions relative to the total read length (if < threshold) is aligned preferentially
Algorithm Options
  -t Int       number of bwa threads [1]
  -k Int       minimum seed length [19]
  -w Int       band width for banded alignment [100]
  -d Int       off-diagonal X drop off [100]
  -r Float     look for internal seeds inside a seed longer than {-k} * FLOAT [1.5]
  -y Int       seed occurrence for the 3rd round seeding [20]
  -c Int       skip seeds with more than INT occurrences [500]
  -D Float     drop chains shorter than FLOAT fraction of the longest overlapping chain [0.50]
  -W Int       discard a chain if seeded bases shorter than INT [0]
  -m Int       perform at most INT rounds of mate rescues for each read [50]
  -S           skip mate rescue
  -P           skip pairing; mate rescue performed unless -S also in use
  -j           ignore ALT contigs
  -T Int       minimum score to output [10]
  -M           mark shorter split hits as secondary
  -I Fl,Fl,Int,Int    specify the mean, standard deviation (10 percent of the mean if absent), 
                      max (4 sigma from the mean if absent) and min of the insert size distribution. 
                      FR orientation only. [inferred]

'''

index_help = '''
BSBolt Index -G {fasta reference} -DB {database output}

-h, --help     show this help message and exit

Index Options:
  -G File      path to reference genome fasta file, fasta file should contain all contigs
  -DB File     path to index directory, alignment index generated inside existing directory 
               or new directory made if directory doesn't exist
  -B Int       block size for bwtsw algorithm, 
               increasing block will speed up indexing and increase memory consumption [10000000]
  -MR File     path to bed file of mappable regions. 
               Index will be built using using masked contig sequence [null]
  -IA          ignore alt contigs when constructing alignment index
  -rrbs        generate a Reduced Representative Bisulfite Sequencing (RRBS) index
  -rrbs-cut-format Str    Cut format to use for generation of RRBS database, [C-CGG] MSPI, 
                          input multiple enzymes as a comma separate string, C-CGG,C-CGG,...
  -rrbs-lower Int      lower bound fragment size to consider RRBS index generation [40]
  -rrbs-upper Int      upper bound fragment size to consider RRBS index generation [500]
'''

meth_help = '''
BSBolt Module CallMethylation -I {input.bam} -DB {BSBolt DB} -O {output prefix}

-h, --help     show this help message and exit

Input / Output Options:
  -I File         input BAM, input file must be in BAM format with index file
  -DB File        path to index directory
  -O File         output prefix
  -text           output plain text files [False]
  -BG             output calls in bedGraph format [False]
  -CG             only output CpG sites in CGmap file [False]
Algorithm Options:
  -remove-ccgg    remove methylation calls in ccgg sites [False]
  -verbose        verbose Output [False]
  -ignore-ov      only consider higher quality base when paired end reads overlap [True]
  -max Int        max read depth to call methylation [8000]
  -min Int        minimum read depth required to report methylation site [10]
  -t Int          methylation calling threads [1]
  -BQ Int         minimum base quality [10]
  -MQ Int         minimum alignment quality [20]
  -IO             ignore orphans reads, (not proper read pair)
'''

aggregate_help = '''
BSBolt AggregateMatrix -F {file1.CGmap,file2.CGmap,...} -O {output_matrix.txt}

-h, --help              show this help message and exit

Options:
  -F File,File,.        comma separated list of CGmap files, 
                        or path to text file with list of line separated CGmap files
  -S Str,Str,.          comma separated list of samples labels. If sample labels are not provided sample labels 
                        are extracted from CGmap files. Can also pass path to txt for line separated sample labels.
  -min-coverage Int     minimum site read depth coverage 
  -min-sample Float     proportion of samples that must have a valid site (above minimum coverage threshold)
  -O File               Aggregate matrix output path
  -CG                   Only output CG sites
  -verbose              Verbose aggregation
  -t Int                Number of threads to use when assembling matrix
  -count                Output a count matrix with count of methylated cytosines and total observed cytosines
'''

sim_help = '''
BSBolt Simulate -G {genome.fa} -O {output_directory}

-h, --help  show this help message and exit

Input / Output Options:
  -G File     path for reference genome fasta file
  -O File     output prefix
  -CG File    path to CGmap file reference profile [Null]
  -overwrite  overwrite previously generated simulation database
  -BR File    Path to previously generated BSBolt methylation reference (directory)
  -NS         don't output simulated methylation counts
  -verbose    verbose read simulation
Algorithm Options:
  -PE         simulate Paired End Reads, default Single End
  -RL Int     simulated Read Length [125]
  -RD Int     simulated Read Depth [20]
  -U          simulate undirectional reads, (bisulfite converted reference strands and PCR products)
  -MR Float   mutation rate [0.005]
  -MI Float   mutation indel fraction [0.20]
  -ME Float   mutation indel extension probability [0.20]
  -RS Int     random seed for variant generation [-1]
  -HA         haplotype mode, homozygous variants only
  -CH         skip simulation of CH methylation, all CH sites unmethylated
  -SE Float   sequencing Error [0.001]
  -NF Float   cutoff threshold for ambiguous bases, simulated reads with a proportion of ambiguous 
              bases above this threshold will not be output [0.05]
  -FM Int     max fragment size [400]
  -IM Int     insert length mean [50]
  -SM Int     insert length standard deviation [50]
'''

impute_help = '''
BSBolt Impute -M {BSBolt_matrix.txt} -O {imputed_matrix.txt}

-h, --help  show this help message and exit

Options:
  -M File     path to BSBolt matrix file
  -B Int      imputation sample batch size kNN imputation, by default the all of the samples 
              will be processed as a single batch
  -W Int      sliding window size for imputation [3000000]
  -k Int      number of neighbors to use for imputation [5]
  -t Int      number of threads available for imputation [1]
  -verbose    verbose imputation
  -O File     output path for imputed matrix
  -R          randomize batches
'''


from collections import defaultdict
from typing import Dict, List, Tuple
import numpy as np
import pysam
from tqdm import tqdm
from BSBolt.Utils.FastqIterator import OpenFastq


def open_alignment_file(alignment_file_path: str = None):
    if alignment_file_path.endswith('.sam'):
        alignment_file = pysam.AlignmentFile(alignment_file_path, 'r')
    else:
        alignment_file = pysam.AlignmentFile(alignment_file_path, 'rb')
    for aligned_segment in alignment_file.fetch(until_eof=True):
        yield aligned_segment


def parse_alignment_comment(read_comment):
    chrom, start, end, cigar, ref_strand = read_comment.replace('+', '').strip().split(':')
    return dict(chrom=chrom, start=int(start),
                end=int(end), cigar=cigar, ref_strand=ref_strand)


def get_read_reference_info(fastq_files: List[str] = None):
    reference_info = {}
    fastq_iterators = (OpenFastq(file) for file in fastq_files)
    for fastq in fastq_iterators:
        for line in fastq:
            read_name = line[0].replace('@', '').strip()
            read_info = parse_alignment_comment(line[2].strip())
            reference_info[read_name] = read_info
    return reference_info


class AlignmentEvaluator:
    """Evaluate alignment against simulated bisulfite sequencing data.

    Params:

    * *duplicated_regions (dict)*: regions duplicated in the simulation reference, [None]
    * *matching_target_prop (float)*: proportion of alignment that most overlap with target region
                                     for a valid alignment to be called, [0.95]
    """

    def __init__(self, duplicated_regions: Dict[str, Tuple[int, int]] = None, matching_target_prop: float = 0.95,
                 verbose: bool = False):
        self.duplicated_regions = duplicated_regions
        self.matching_target_prop = matching_target_prop
        self.verbose = verbose

    def evaluate_alignment(self, alignment_file: str, fastq_files: List[str] = None) -> Dict[str, int]:
        """

        Params:

        * *alignment_file (str)*: path to alignment file
        * *fastq_files (list)*: list of paths to fastq files

        Returns:

        * *alignment_evaluations (dict)*: target alignment stats
        """
        alignment_evaluations = defaultdict(int)
        reads_observed = set()
        reference_info = get_read_reference_info(fastq_files=fastq_files)
        for alignment in tqdm(open_alignment_file(alignment_file), disable=True if not self.verbose else False):
            # read shouldn't have read pair info but added for safety and tools that don't remove
            read_name = alignment.qname.split('/')[0]
            if alignment.is_read1 or not alignment.is_paired:
                read_name = f'{read_name}/1'
            else:
                read_name = f'{read_name}/2'
            reads_observed.add(read_name)
            alignment_evaluations['ObservedAlignments'] += 1
            alignment_info = reference_info[read_name]
            if not alignment.is_unmapped:
                dup_region = False
                if alignment_info['chrom'] in self.duplicated_regions:
                    duplicate_range = self.duplicated_regions[alignment_info['chrom']]
                    if duplicate_range[0] <= alignment_info['start'] <= duplicate_range[1]:
                        dup_region = True
                chrom_match, matching_prop = self.assess_alignment(alignment, alignment_info)
                on_target = 'On' if chrom_match and matching_prop >= self.matching_target_prop else 'Off'
                secondary = 'Sec' if alignment.is_secondary else 'Prim'
                pair = 'PropPair' if alignment.is_proper_pair else 'Discord'
                dup = 'Dup' if dup_region else 'NoDup'
                alignment_evaluations[f'{on_target}_{secondary}_{pair}_{dup}'] += 1
            else:
                alignment_evaluations['UnalignedAlignments'] += 1
            alignment_evaluations['ReadObserved'] = len(reads_observed)
            alignment_evaluations['SimulatedReads'] = len(reference_info.keys())
        return alignment_evaluations

    @staticmethod
    def assess_alignment(alignment: pysam.AlignedSegment, alignment_info: Dict):
        """ Compare alignment against reference alignment"""
        chrom_match = alignment.reference_name == alignment_info['chrom']
        # assess reference bases that match between the two reads
        matching_pos = np.array(alignment.get_reference_positions(full_length=False))
        base_range = (matching_pos >= alignment_info['start']) & (matching_pos <= alignment_info['end'])
        matching_prop = sum(base_range) / len(alignment_info['cigar'])
        return chrom_match, matching_prop



import re
from typing import List, Tuple

from BSBolt.Index.RRBSCutSites import ProcessCutSites
from BSBolt.Index.IndexOutput import IndexOutput
from BSBolt.Utils.FastaIterator import OpenFasta
from BSBolt.Utils.UtilityFunctions import get_external_paths


class RRBSBuild:
    """Format reference sequence inputs for processing by BWA. In silico digests reference sequence and
    return mappable regions that are within the fragment boundary. Fragments are relative to the restriction cut site
    if provided, or the complete restriction sequence is considered as part of the mappable fragment.:

    Params:

    * *reference_file (str)*: path to reference file in fasta format
    * *genome_database (str)*: directory to output processed datafiles
    * *block_size (int)*: bwa indexing block size, increases indexing speed but increases memory consumption
    * *lower_bound (int)*: smallest mappable fragment size
    * *upper_bound (int)*: largest mappable fragment size
    * *cut_format (str)*: Comma separated list of restriction sites, - represent cut break
    * *ignore_alt (bool)*: ignore alt contigs when constructing alignment index

    Usage:
    ```python
    index = RRBSBuild(**kwargs)
    index.generate_rrbs_database()
    ```
        """

    def __init__(self, reference_file: str = None, genome_database: str = None,
                 lower_bound: int = 30, upper_bound: int = 500, cut_format: str = 'C-CGG',
                 block_size: int = None, ignore_alt: bool = False):
        bwa_path, _, _ = get_external_paths()
        self.reference_file = OpenFasta(fasta=reference_file)
        self.index_output = IndexOutput(genome_database=genome_database, block_size=block_size)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.ignore_alt = ignore_alt
        self.cut_sites = ProcessCutSites(cut_format=cut_format)
        self.mappable_regions = []
        self.contig_size_dict = {}

    def generate_rrbs_database(self):
        """ Wrapper for class functions to process and build mapping indices.
        """
        contig_id = None
        contig_sequence = []
        # return true, line if '>' in fasta line
        for is_contig_label, sequence in self.reference_file:
            if is_contig_label:
                if contig_id:
                    self.process_contig_region(contig_id, contig_sequence)
                    contig_sequence = []
                # set contig_id
                contig_id = sequence.replace('>', '').split()[0]
            else:
                contig_sequence.append(sequence.upper())
        # process remaining sequence after iterator exhausted
        self.process_contig_region(contig_id, contig_sequence)
        self.index_output.database_output.close()
        # launch build commands
        self.index_output.build_index()
        # output mappable regions in bed format
        self.index_output.output_mappable_regions(self.mappable_regions)
        # output contig size index
        self.index_output.output_contig_sequence('genome_index', self.contig_size_dict)

    def process_contig_region(self, contig_id: str, contig_sequence: List[str]):
        """Given a contig_id will output a pickle file of the whole sequence and output a masked version of the
        the sequence where only mappable regions are reported.

        Params:

        * *contig_id (str)*: contig label
        * *contig_sequence (list)*: a list of of string containing DNA Sequence
        """
        if self.ignore_alt and 'alt' in contig_id.lower():
            return
        # join contig_squence to get ease downstream processing
        contig_str: str = ''.join(contig_sequence)
        # save contig size to dict
        self.contig_size_dict[contig_id] = len(contig_str)
        # serialize contig sequence
        self.index_output.output_contig_sequence(contig_id, contig_str)
        # retrieve list of mappable regions
        mappable_regions: list = self.process_rrbs_sequence(contig_str)
        # check if contig regions empty, if so unmask first 80 bases for indexing
        if not mappable_regions:
            mappable_regions.append((1, 80))
        # get masked contig sequence
        masked_contig_sequence: str = self.mask_contig(contig_str, mappable_regions)
        # perform sanity check, if AssertionError the region designation process is bad
        assert len(contig_str) == len(masked_contig_sequence), 'Contig Length != Masked Contig Length'
        # write contig sequence to output files
        self.index_output.write_contig_sequence(contig_id, masked_contig_sequence)

    def process_rrbs_sequence(self, contig_str: str) -> List[Tuple[int, int]]:
        """Designate mappable regions by finding all occurrences of the restriction site string in the passed DNA
        sequence. Merge restriction map into regions by considering pairs of downstream and upstream restriction sites
        that pass the size limits.

        Params:

        * *contig_str (str)*: STR of continuous DNA sequence

        Returns:

        * *mappable_regions (list)*: List of tuples the contain the start and end position of fragments that
            are with the size limits
        """
        restriction_site_locations = []
        # get the position of the all occurrences of the restriction site pattern in the DNA sequence and add to list
        for restrication_site, offset in self.cut_sites.restriction_site_dict.items():
            restriction_site_locations.extend([(m.start(), offset, restrication_site) for m in
                                               re.finditer(restrication_site, contig_str)])
        # sort list so fragments are in order or start position
        restriction_site_locations.sort(key=lambda x: x[0])
        mappable_regions = []
        # iterate through list to get the cut site plus upstream cut site, will terminate at the second to last site
        for downstream_site, upstream_site in zip(restriction_site_locations, restriction_site_locations[1:]):
            # set default downstream site that includes the restriction site
            down_start = downstream_site[0]
            # if cut site indicated shift start position
            if downstream_site[1]:
                down_start = downstream_site[0] + downstream_site[1]
            # set site end to include restriction site
            up_end = upstream_site[0] + len(upstream_site[2])
            # correct fragment if cut site provided
            if upstream_site[1]:
                up_end = upstream_site[0] + upstream_site[1]
            site_range = up_end - down_start
            # check fragment length
            if self.lower_bound <= site_range <= self.upper_bound:
                # extend mappable boundaries by two to assess methylation context relative to reference sequence
                mappable_regions.append((down_start - 2, up_end + 2))
        return mappable_regions

    def mask_contig(self, contig_str: str, mappable_regions: List[Tuple[int, int]]) -> str:
        """Given a list of mappable regions, if cut site isn't designated merges mappable fragments, returns a string
                of DNA sequence with masked unmappable regions

                Params:

                * *contig_id (str)*: contig label
                * *contig_str (str)*: str of DNA sequence
                * *mappable_regions (list)*: list of mappable regions

                Returns:

                * *masked_contig_sequence (str)*: str of DNA sequence with un-mappable regions masked
        """
        contig_mappable_regions = iter(mappable_regions)
        start, end = next(contig_mappable_regions)
        masked_contig = []
        for position, bp in enumerate(contig_str):
            if position > end:
                try:
                    start, end = next(contig_mappable_regions)
                except StopIteration:
                    pass
            if start <= position <= end:
                masked_contig.append(bp)
            else:
                masked_contig.append('-')
        return ''.join(masked_contig)



import random
import numpy as np
from tqdm import tqdm


class MaskImputationValues:
    """
    Mask known beta values from input pd.DataFrame to evaluate imputation accuracy
        Keyword Arguments
            methylation_array (np.array): input pd.DataFrame of methylation values
            masking_proportion (float, list): proportion of sites to mask, per sample or overall
            masking_sites (dict): dict of sites that were masked hashing to known value
            verbose (bool): use tqdm progress bar or suppress
        Attributes:
            self.methylation_dataframe (np.array): input pd.DataFrame of methylation values
            self.masking_proportion (float, list): proportion of sites to mask, per sample or overall
            self.masking_sites (dict): dict of sites that were masked hashing to known value
            self.disable_tqdm (bool): if verbose False else True
    """

    def __init__(self, methylation_array=None, masking_proportion=None, masking_sites=None, verbose=False):

        self.methylation_array = np.copy(methylation_array)
        self.masking_proportion = masking_proportion
        self.masking_sites = masking_sites
        if self.masking_sites:
            self.masking_sites = dict(self.masking_sites)
        self.tqdm_disable = True
        if verbose:
            self.tqdm_disable = False

    def get_masking_proportion(self, sample_index):
        """ Return masking probability for site, return self.masking_probabiliyt if float given else return
        item at self.masking_proportion[sample_index]"""
        if isinstance(self.masking_proportion, float):
            return self.masking_proportion
        else:
            assert len(self.methylation_array[0]) == len(self.masking_proportion), 'Masking proportion mush be ' \
                                                                                   'provided for every sample'
            return self.masking_proportion[sample_index]

    @staticmethod
    def mask_value(masking_proportion):
        """Return True if random.random() less than masking proportion"""
        return random.random() < masking_proportion

    def mask_random_sites(self):
        """Mask random sites if called based on masking_proportion provided"""
        # ensure self.masking_sites is empty if called
        self.masking_sites = {}
        # iterate through rows
        for row_index, row_values in tqdm(enumerate(self.methylation_array), total=self.methylation_array.shape[0],
                                          disable=self.tqdm_disable, desc='Masking Sites'):
            for sample_index, sample_value in enumerate(row_values):
                masking_proportion = self.get_masking_proportion(sample_index)
                # if true mask site
                if self.mask_value(masking_proportion):
                    # update known value
                    self.masking_sites[f'{row_index}_{sample_index}'] = sample_value
                    row_values[sample_index] = np.nan

    def mask_known_sites(self):
        """Mask sites using provided dict of previously masked sites"""
        for row_index, row_values in tqdm(enumerate(self.methylation_array), total=self.methylation_array.shape[0],
                                          disable=self.tqdm_disable, desc='Masking Sites'):
            for sample_index, sample_value in enumerate(row_values):
                if f'{row_index}_{sample_index}' in self.masking_sites:
                    row_values[sample_index] = np.nan



import gzip
import io
import random
from typing import List, Tuple
import numpy as np
from BSBolt.Impute.Imputation.GenomeImputation import GenomeImputation
from BSBolt.Impute.Impute_Utils.ImputationFunctions import get_bsb_matrix


class ImputeMissingValues:
    """
    Launch and knn imputation task. This wrapper imports data for imputation, split data for batch imputation,
    and combines data after imputation. Data is held in memory for access during imputation. If closest neighbors null,
    null imputed value returned.

    Params:

    * *input_matrix_file (str)*: Path to BSBolt matrix
    * *batch_size (int)*: Batch size for batch imputation
    * *imputation_window_size (int)*: Size (bp) for imputation window, [3,000,000]
    * *k (int)*: Nearest neighbors used for imputation, [5]
    * *threads (int)*: Number of threads available for imputation, [1]
    * *verbose (bool)*: Verbose imputation, [False]
    * *sep (str)*: separator character used in methylation matrix, [\t]
    * *output_path (str)*: output path
    * *randomize_batch (bool)*: randomize batch, [False]
    * *meth_matrix (np.ndarray)*: imputed methylation matrix
    * *meth_site_order (list)*: ordered methylation sites, sorted by contig then position
    * *sample_ids (list)*: sample names
    """

    def __init__(self, input_matrix_file: str = None, batch_size: int = None, imputation_window_size: int = 3000000,
                 k: int = 5, threads: int = 4, verbose: bool = False, sep: str = '\t', output_path: str = None,
                 randomize_batch: bool = False, meth_matrix: np.ndarray = None, meth_site_order: list = None,
                 sample_ids: list = None):
        self.input_matrix_file = input_matrix_file
        self.meth_matrix = meth_matrix
        self.meth_site_order = meth_site_order
        self.sample_ids = sample_ids
        self.batch_commands = [0, False]
        if batch_size:
            self.batch_commands = [batch_size, randomize_batch]
        self.output_matrix = None
        if output_path:
            self.output_matrix = self.get_output_matrix(output_path)
        self.tqdm_disable = False if verbose else True
        self.sep = sep
        self.imputation_kwargs = dict(imputation_window_size=imputation_window_size,
                                      k=k,
                                      threads=threads,
                                      verbose=verbose)

    def impute_values(self):
        """
        Launch kNN imputation for each batch and set values in original matrix.
        """
        imputation_order = [count for count in range(len(self.sample_ids[1]))]
        if self.batch_commands[1]:
            random.shuffle(imputation_order)
        if not self.batch_commands[0]:
            self.batch_commands[0] = len(self.sample_ids[1])
        imputation_batches = self.process_batch(imputation_order)
        for batch in imputation_batches:
            batch_array, sample_labels = self.get_batch_data(batch)
            batch_impute = self.launch_genome_imputation(batch_array, sample_labels)
            for count, sample in enumerate(batch):
                self.meth_matrix[:, sample] = batch_impute.genomic_array[:, count]

    def process_batch(self, imputation_order: List[int]) -> List[List[int]]:
        """Generate sample batches"""
        batches = []
        if not self.batch_commands[0]:
            self.batch_commands[0] = len(self.sample_ids)
        batch_number = int(len(imputation_order) / self.batch_commands[0])
        batch_remainder = len(imputation_order) % self.batch_commands[0]
        if float(batch_remainder) / float(self.batch_commands[0]) <= .6 and batch_remainder != 0:
            batch_addition = int(batch_remainder / batch_number)
            self.batch_commands[0] += batch_addition + 1
            print(f'Adjusting batch size, new batch size = {self.batch_commands[0]}')
        start, end = 0, self.batch_commands[0]
        while True:
            batch_order = imputation_order[start: end]
            if not batch_order:
                break
            batches.append(batch_order)
            start += self.batch_commands[0]
            end += self.batch_commands[0]
        return batches

    def get_batch_data(self, batch: List[int]) -> Tuple[np.ndarray, List[str]]:
        """Return methylation value for batch imputation

        Returns:

        * *batch_array (np.ndarry)*: array of methylation values
        * *sample_labels (list)*: list of samples in batch
            """
        batch_array = self.meth_matrix[:, batch]
        sample_labels = [self.sample_ids[1][sample] for sample in batch]
        return batch_array, sample_labels

    def import_matrix(self):
        self.meth_matrix, self.meth_site_order, self.sample_ids = get_bsb_matrix(self.input_matrix_file)

    @staticmethod
    def get_output_matrix(output_path: str):
        """Get output object"""
        if output_path.endswith('.gz'):
            out = io.BufferedWriter(gzip.open(output_path, 'wb'))
        else:
            out = open(output_path, 'w')
        return out

    def output_imputed_matrix(self):
        """Write imputed values"""
        self.output_matrix.write('\t'.join([self.sample_ids[0]] + self.sample_ids[1]) + '\n')
        for site_label, values in zip(self.meth_site_order, self.meth_matrix):
            str_values = '\t'.join([str(value) for value in values])
            self.output_matrix.write(f'{site_label}\t{str_values}\n')
        self.output_matrix.close()

    def launch_genome_imputation(self, meth_array: np.ndarray, sample_labels: List) -> object:
        imputation_kwargs = dict(self.imputation_kwargs)
        imputation_kwargs.update(dict(genomic_array=meth_array,
                                      sample_labels=sample_labels,
                                      row_labels=self.meth_site_order))
        knn_imputation: GenomeImputation = GenomeImputation(**imputation_kwargs)
        knn_imputation.impute_windows()
        return knn_imputation



import random
from typing import Dict, List
import numpy as np
from BSBolt.Utils.FastaIterator import OpenFasta
from BSBolt.Utils.CGmapIterator import OpenCGmap
from BSBolt.Simulate.SimulationOutput import SimulationOutput


class SetCytosineMethylation:
    """Set cytosine methylation at CpG and CH sites. Methylation value are assigned for at + and - cytosines. After
    setting a methylation value, a random roll is performed for each read. If the roll is less than the methylation
    value the cytosine is set as methylation and if it is greater than the cytosine is unmethylated. This introduces
    some random noise into the simulation process. The number of methylated and unmethylated reads is tracked in the
    final simulation output.
    Keyword Arguments:
        reference_file (str): path to reference genome
        methylation_reference_output (str): path to output folder for simulation data
        methylation_reference (str): CGmap file with methylation profile
        methylation_profile (str): path to previously generated methylation reference
    Attributes:
        self.reference_file (str): path to reference genome
        self.methylation_reference_output (str): path to output folder for simulation data
        self.methylation_reference (str): CGmap file with methylation profile
        self.methylation_profile (str): path to previously generated methylation reference
        self.reference_dict (dict): contig sequences
        self.reference_contig_size (dict): lens of all contigs
        self.methylation_profile (dict): dict with cytosine methylation values
        self.simulation_out (dict): objects for loading and outputting simulation data
        self.cpg_distribution (np.array): beta binomial distribution for CpG sites
        self.ch_distribution (np.array): beat binomial distribution for CH sites
            """

    def __init__(self, reference_file: str = None, sim_dir: str = None, methylation_reference: str = None,
                 cgmap: str = None, collect_ch_sites: bool = True, overwrite_db: bool = False):
        self.reference = self.get_reference(reference_file)
        self.reference_file = reference_file
        self.sim_dir = sim_dir
        self.sim_db: SimulationOutput = None
        self.initialize_methylation_reference(methylation_reference, cgmap)
        self.methylation_reference = methylation_reference
        self.collect_ch_sites = collect_ch_sites
        self.overwrite_db = overwrite_db
        self.cpg_distribution = np.random.beta(.5, .5, size=5000)
        self.ch_distribution = np.random.beta(.01, .05, size=5000)

    def initialize_methylation_reference(self, methylation_reference, cgmap):
        """If profile exists then assume already run"""
        if methylation_reference:
            meth_in = SimulationOutput(sim_dir=methylation_reference)
            meth_out = SimulationOutput(sim_dir=self.sim_dir)
            meth_out.generate_simulation_directory()
            for contig in self.reference.keys():
                for contig_key, contig_values in meth_in.load_contig(contig):
                    if contig_key:
                        meth_out.output_contig(contig_key, contig_values, contig)
            self.sim_db = meth_out
        elif cgmap:
            self.sim_db = SimulationOutput(sim_dir=self.sim_dir)
            self.sim_db.generate_simulation_directory()
            self.sim_db_from_cgmap(cgmap)
        else:
            self.sim_db = SimulationOutput(sim_dir=self.sim_dir)
            self.sim_db.generate_simulation_directory()

    def sim_db_from_cgmap(self, cgmap: str):
        meth_keys = {}
        cytosine_values = {}
        cytosine_positions = {}
        for line in OpenCGmap(cgmap):
            chrom, nucleotide, pos, context, methlevel = line[0], line[1], int(line[2]) - 1, line[4], float(line[7])
            context = 1 if context == 'CG' else 0
            if not self.collect_ch_sites and not context:
                continue
            nucleotide = 1 if nucleotide == 'G' else 0
            # nucleotide, methylation_level, context, methylated_reads, unmethylated_reads
            meth_profile = np.assarray([nucleotide, methlevel, context, 0, 0, 1])
            profile_id = f'{chrom}:{pos}'
            if chrom not in meth_keys:
                meth_keys[chrom] = {profile_id: 0}
                cytosine_values[chrom] = [meth_profile]
                cytosine_positions[chrom] = 1
            else:
                meth_keys[chrom][profile_id] = cytosine_positions[chrom]
                cytosine_values[chrom].append(meth_profile)
                cytosine_positions[chrom] += 1
        for contig, meth_key in meth_keys.items():
            self.sim_db.output_contig(meth_key, np.array(cytosine_values[contig]), contig)

    @property
    def pick_cpg_methylation(self):
        """Sample from CpG distribution"""
        return random.choice(self.cpg_distribution)

    @property
    def pick_ch_methylation(self):
        """Sample from CH distribution"""
        return random.choice(self.ch_distribution)

    @staticmethod
    def get_reference(reference_file: str) -> Dict[str, str]:
        """Retrieve contig sequence from reference fasta rather than index to ensure simulation will work without
        a generated index. """
        # store contig_id and matching sequencing
        reference = {}
        contig_id = None
        contig_sequence = []
        for contig_label, line in OpenFasta(reference_file):
            if contig_label:
                # join sequence and add to dict if new contig encountered
                if contig_id:
                    reference[contig_id] = ''.join(contig_sequence)
                    contig_sequence = []
                contig_id = line.replace('>', '').split()[0]
            else:
                contig_sequence.append(line.upper())
        # join contig and add to dict after iteration finishes
        reference[contig_id] = ''.join(contig_sequence)
        return reference

    def get_contig_methylation(self, contig):
        if self.overwrite_db:
            contig_profile = None
        else:
            contig_profile = self.sim_db.load_contig(contig)
        if contig_profile:
            return contig_profile
        else:
            return self.set_random_cytosine_methylation(contig)

    def set_variant_methylation(self, sim_data, contig_profile, current_contig):
        """Variants are always random, so set random methylation"""
        ref_seq = self.reference[current_contig]
        for pos, variant_info in sim_data.items():
            # don't set methylation for last base
            if pos + 2 > len(ref_seq):
                continue
            if variant_info['indel'] == -1:
                continue
            elif variant_info['indel'] == 1:
                insert_context = f'{ref_seq[pos]}{variant_info["alt"]}{ref_seq[pos + 1]}'
                for insert_pos, nucleotide in enumerate(variant_info['alt']):
                    local_context = insert_context[insert_pos: insert_pos + 3]
                    meth_profile = self.get_random_methylation_profile(nucleotide, local_context)
                    if meth_profile:
                        contig_profile[f'{pos}_+_{insert_pos + 1}'] = meth_profile
            else:
                for nucleotide in variant_info['iupac']:
                    if nucleotide == variant_info['reference']:
                        continue
                    else:
                        context = f'{ref_seq[pos - 1]}{nucleotide}{ref_seq[pos + 1]}'
                        meth_profile = self.get_random_methylation_profile(nucleotide, context)
                        if meth_profile:
                            contig_profile[f'{pos}_{variant_info["reference"]}_{nucleotide}'] = meth_profile

    def set_random_cytosine_methylation(self, contig):
        """Iterate through reference sequence, upon encountering a Cytosine (Watson) or Guanine (Crick):
            1. get nucleotide context
            2. pull random methylation value from appropriate distribution and store value
        """
        # iterate through reference sequence
        sequence = self.reference[contig]
        cytosine_profile = {}
        for pos, nucleotide in enumerate(sequence):
            # retrieve nucleotide context
            context = sequence[pos - 1: pos + 2]
            # context won't exist at ends of reference sequence
            if context:
                # set watson methylation
                methylation_profile = self.get_random_methylation_profile(nucleotide, context)
                if methylation_profile:
                    # if no CH information don't add info
                    if not self.collect_ch_sites and not methylation_profile[2]:
                        continue
                    cytosine_profile[f'{pos}'] = tuple(methylation_profile)
        return cytosine_profile

    def get_random_methylation_profile(self, nucleotide, context):
        if nucleotide == 'C':
            n_context = 1 if context[1:] == 'CG' else 0
            return self.get_methylation_level(n_context, 0)
        # set crick methylation
        elif nucleotide == 'G':
            n_context = 1 if context[0:2] == 'CG' else 0
            return self.get_methylation_level(n_context, 1)
        return None

    def get_methylation_level(self, context, nucleotide) -> tuple:
        """Retrieve methylation level from distribution based on nucleotide context
        Arguments:
            context (int): 2BP nucleotide context
            nucleotide (int): reference nucleotide
        Return:
            methylation_profile (np.ndarray): context, methylation level, unpopulated methylation read information
            """
        if context:
            methylation_level = self.pick_cpg_methylation
        else:
            methylation_level = self.pick_ch_methylation
        return nucleotide, methylation_level, context



#! /usr/bin/env python3

import sys
from BSBolt.Utils.UtilityFunctions import check_python_version, check_package_version
from BSBolt.Utils.Launcher import bsb_launch
from BSBolt.Utils.Parser import parser


def launch_bsb():
    check_python_version()
    if not check_package_version():
        print('Please Update Packages')
        sys.exit()

    arguments = parser.parse_args()

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        # parser.print_usage() # for just the usage line
        parser.exit()

    launcher = bsb_launch[arguments.subparser_name]
    launcher(arguments)


if __name__ == "__main__":
    launch_bsb()



from collections import Counter
import pickle
from typing import Dict, Tuple, Union
import pysam


class CallMethylationValues:
    """
    Params:

    * *input_file (str)*: str to input bam/sam file
    * *genome_database (str)*: str to genome directory
    * *ignore_overlap (bool)*:  ignore overlapping signal, calls methylation using the higher quality base for
                                reads with overlapping alignments, [True]
    * *remove_ccgg (bool)*: don't call CCGG sequences, [False]
    * *ignore_oprhans (bool)*: ignore orphaned reads (not properly paired), [True]
    * *max_read_depth (int)*: maximum read depth for pileup, [8000]
    * *contig (str)*: contig to process
    * *min_base_quality (int)*: minimum quality for a base to be reported for methylation calling, [10]
    * *return_queue (Queue.queue)*: results are added to a queue in batches for multi-threaded access
    * *cg_only (bool)*: only return CG sites to queue, [False]
    * *min_mapping_quality (int)*: minimum mapping quality for an alignment to be used for methylation calling, [10]

    Attributes:

    * *self.chunk_size (int)*: Number of sites in chunk added to queue, [10000]

    """

    def __init__(self, input_file: str = None, genome_database: str = None,
                 ignore_overlap: bool = True, remove_ccgg: bool = False, ignore_orphans: bool = True,
                 max_read_depth: int = 8000, contig: str = None, min_base_quality: int = 10, return_queue=None,
                 cg_only: bool = False, min_mapping_quality: int = 10):
        self.input_file = str(input_file)
        self.input_bam = pysam.AlignmentFile(self.input_file, mode='rb',
                                             require_index=True)
        self.genome_database = str(genome_database)
        if self.genome_database[-1] != '/':
            self.genome_database = f'{self.genome_database}/'
        self.ignore_overlap = ignore_overlap
        self.remove_ccgg = remove_ccgg
        self.ignore_orphans = ignore_orphans
        self.max_read_depth = max_read_depth
        self.contig = contig
        self.min_base_quality = min_base_quality
        self.chunk_size = 10000
        self.cg_only = cg_only
        self.min_mapping_quality = min_mapping_quality
        self.context_tables = self.get_context_tables
        self.return_queue = return_queue
        self.counting_dict = {}

    @property
    def get_context_tables(self) -> Dict[str, Dict[str, str]]:
        """
        Returns:

        * *context_tables (dict)*: base methylation context relative to sense strand
        """
        context_tables = {'context_table': {'CAA': 'CHH', 'CAC': 'CHH', 'CAG': 'CHG', 'CAT': 'CHH', 'CCA': 'CHH',
                                            'CCC': 'CHH', 'CCG': 'CHG', 'CCT': 'CHH', 'CGA': 'CG', 'CGC': 'CG',
                                            'CGG': 'CG', 'CGT': 'CG', 'CTA': 'CHH', 'CTC': 'CHH', 'CTG': 'CHG',
                                            'CTT': 'CHH'},
                          'sub_context_table': {'CAA': 'CA', 'CAC': 'CA', 'CAG': 'CA', 'CAT': 'CA', 'CCA': 'CC',
                                                'CCC': 'CC', 'CCG': 'CC', 'CCT': 'CC', 'CGA': 'CG', 'CGC': 'CG',
                                                'CGG': 'CG', 'CGT': 'CG', 'CTA': 'CT', 'CTC': 'CT', 'CTG': 'CT',
                                                'CTT': 'CT'},
                          'antisense_context_table': {'TTG': 'CHH', 'TGG': 'CHH', 'TCG': 'CG', 'TAG': 'CHH',
                                                      'GTG': 'CHH', 'GGG': 'CHH', 'GCG': 'CG', 'GAG': 'CHH',
                                                      'CTG': 'CHG', 'CGG': 'CHG', 'CCG': 'CG', 'CAG': 'CHG',
                                                      'ATG': 'CHH', 'AGG': 'CHH', 'ACG': 'CG', 'AAG': 'CHH'},
                          'antisense_sub_context_table': {'TTG': 'CA', 'TGG': 'CC', 'TCG': 'CG', 'TAG': 'CT',
                                                          'GTG': 'CA', 'GGG': 'CC', 'GCG': 'CG', 'GAG': 'CT',
                                                          'CTG': 'CA', 'CGG': 'CC', 'CCG': 'CG', 'CAG': 'CT',
                                                          'ATG': 'CA', 'AGG': 'CC', 'ACG': 'CG', 'AAG': 'CT'}}
        return context_tables

    def call_methylation(self):
        """Run methylation call for contig"""
        try:
            chrom_seq = self.get_reference_sequence(f'{self.genome_database}{self.contig}.pkl')
        except FileNotFoundError:
            print(f'{self.contig} not found in BSBolt DB, Methylation Calls for {self.contig} skipped. Methylation '
                  f'values should be called using the same DB used for alignment.')
            self.return_queue.put([])
        else:
            self.call_contig(chrom_seq)

    def call_contig(self, chrom_seq: str):
        """Iterates through bam pileup, calling methylation values if the reference nucleotide is a C or G. Pileup reads
        are buffered and accessed as needed. Process sites are appended to list and returned in chunks.

        Reads flagged as:

        * pcr duplicates (1024), read unmapped (4), read fails platform/vendor quality checks (512) are ignored
        """
        # iterate through pileup
        line_count = 0
        contig_chunk = []
        # flag_require (0) mapped read
        # flag_filter 1540 = pcr duplicates (1024) + read unmapped (4) + read fails platform/vendor quality checks (512)
        for pileup_col in self.input_bam.pileup(max_depth=self.max_read_depth,
                                                contig=self.contig,
                                                ignore_overlaps=self.ignore_overlap,
                                                min_base_quality=self.min_base_quality,
                                                ignore_orphans=self.ignore_orphans,
                                                min_mapping_quality=self.min_mapping_quality,
                                                flag_require=0,
                                                flag_filter=1540):
            # get sequence around pileup site
            reference_seq = chrom_seq[(pileup_col.reference_pos - 3):(pileup_col.reference_pos + 4)].upper()
            # get nucleotide context
            fivemer = reference_seq[1:-1]
            if len(fivemer) == 5:
                nucleotide = fivemer[2]
                context, subcontext = self.get_context(nucleotide, fivemer)
                if self.cg_only and subcontext != 'CG':
                    continue
                # check if CCGG in sequence, skip loop if filter True
                if self.check_ccgg(reference_seq):
                    continue
                # count the pileup read bases, Uppercase watson strand, lowercase crick strand
                try:
                    base_counts = Counter(pileup_col.get_query_sequences(mark_matches=False,
                                                                         mark_ends=False,
                                                                         add_indels=False))
                except AssertionError:
                    # pysam may throw an error if the number of reads in pileup column is above the max
                    continue
                meth_line = self.get_methylation_call(nucleotide, base_counts)
                meth_line.update({'pos': pileup_col.reference_pos + 1, 'chrom': self.contig,
                                  'context': context, 'subcontext': subcontext})
                contig_chunk.append(tuple(meth_line.values()))
                line_count += 1
                # return chunk for output
                if line_count == self.chunk_size - 1:
                    self.return_queue.put(contig_chunk, block=True)
                    contig_chunk = []
                    line_count = 0
        self.return_queue.put(contig_chunk, block=True)

    def get_context(self, nucleotide: str, fivemer: str) -> Tuple[str, str]:
        """
        Params:

        * *nucleotide (str)*: 1 nucleotide
        * *fivemer (str)*: 5 nucleotides

        Returns:

        * *context (str)*: methylation context
        * *subcontext (str)*: methylation subcontext
        """
        null_context = '--'
        context = null_context
        subcontext = null_context
        if nucleotide == 'C':
            subcontext = self.context_tables['sub_context_table'].get(fivemer[2:5], null_context)
            context = self.context_tables['context_table'].get(fivemer[2:5], null_context)
        elif nucleotide == 'G':
            subcontext = self.context_tables['antisense_sub_context_table'].get(fivemer[0:3], null_context)
            context = self.context_tables['antisense_context_table'].get(fivemer[0:3], null_context)
        return context, subcontext

    def check_ccgg(self, sequence: str) -> bool:
        """checks if sequence is == to CCGG
        """
        if self.remove_ccgg:
            return 'CCGG' in sequence
        return False

    @staticmethod
    def get_methylation_call(nucleotide: str, base_counts: Dict[str, int]) -> Dict[str, Union[int, float, str]]:
        """
        Methylation for a C relative to the sense strand of the reference can only be called using watson reads,
        and G with crick reads

        Param:

        * *nucleotide (str)*: reference nucleotide
        * *base_counts (collections.Counter)*: watson nucleotides are Uppercase and crick nucleotides lowercase

        Returns:

        * *meth_line (dict)*: formatted methylation calls for output
        """
        meth_cytosines = 0
        unmeth_cytosines = 0
        # call cytonsines with watson
        if nucleotide == 'C':
            meth_cytosines = base_counts.get('C', 0)
            unmeth_cytosines = base_counts.get('T', 0)
        # call guanines with crick strand
        elif nucleotide == 'G':
            meth_cytosines = base_counts.get('g', 0)
            unmeth_cytosines = base_counts.get('a', 0)
        all_cytosines = meth_cytosines + unmeth_cytosines
        meth_level = 'na'
        if all_cytosines > 0:
            # if cytosines present call methylation
            meth_level = round(float(meth_cytosines) / float(all_cytosines), 6)
        f'{base_counts.get("A", 0)}\t{base_counts.get("T", 0)}\t{base_counts.get("C", 0)}' \
            f'\t{base_counts.get("G", 0)}\t{base_counts.get("N", 0)}'
        forward_counts = f'{base_counts.get("A", 0)}\t{base_counts.get("T", 0)}\t{base_counts.get("C", 0)}' \
            f'\t{base_counts.get("G", 0)}\t{base_counts.get("N", 0)}'
        reverse_counts = f'{base_counts.get("a", 0)}\t{base_counts.get("t", 0)}\t{base_counts.get("c", 0)}' \
            f'\t{base_counts.get("g", 0)}\t{base_counts.get("n", 0)}'
        return {'nucleotide': nucleotide, 'meth_cytosines': meth_cytosines, 'unmeth_cytosines': unmeth_cytosines,
                'all_cytosines': all_cytosines, 'meth_level': meth_level, 'forward_counts': forward_counts,
                'reverse_counts': reverse_counts}

    @staticmethod
    def get_reference_sequence(path: str) -> str:
        """load serialized reference file from path
        """
        with open(path, 'rb') as genome_file:
            return pickle.load(genome_file)



import io
import gzip
import multiprocessing
from typing import Dict, List, Tuple, Union
import pysam
from tqdm import tqdm
from BSBolt.CallMethylation.CallValues import CallMethylationValues


class MethylationCallingError(Exception):
    """Error in methylation calling process"""
    pass


def call_contig_methylation(completed_contigs, call_methylation_kwargs):
    """ Wrapper to initialize methylation calling
    Arguments:
        completed_contigs (multiprocessing.manager.list): List of contigs with completed methylation calls
        call_methylation_kwargs (dict): dict of argument for CallMethylation class
    """
    contig_methylation_call = CallMethylationValues(**call_methylation_kwargs)
    contig_methylation_call.call_methylation()
    assert isinstance(contig_methylation_call, CallMethylationValues)
    completed_contigs.append(call_methylation_kwargs['contig'])


class ProcessContigs:
    """
    Multi-threaded contig processing wrapper. Passes thread safe queue to workers and outputs values as CGmap file.

    Params:

   * *input_file (str)*: str to input bam/sam file
   * *genome_database (str)*: str to genome directory
   * *output_prefix (str)*: output prefix for  CGmap
   * *ignore_overlap (bool)*:  ignore overlapping reads, [True]
   * *text_output (bool)*: output compressed or plain text files, [False]
   * *remove_ccgg (bool)*: don't call CCGG sequences, [False]
   * *min_read_depth (int)*: default = minimum read depth to call methylation values, [1]
   * *max_read_depth (int)*: maximum read depth for pileup, [8000]
   * *threads (int)*: , if one watcher and processing on same thread else separated, [1]
   * *min_base_quality (int)*: minimum base quality for base to be considered, [10]
   * *min_mapping_quality (int)*: minimum mapping quality for an alignment to be considered, [10]
   * *verbose (bool)*: Verbose processing, [False]
   * *cg_only (bool)*: only return CG sites to queue, [False]
   * *ignore_oprhans (bool)*: ignore orphaned reads (not properly paired), [True]
   * *bedgraph_output* (bool)*: output bedgraph format inplace of CGmap, [False]

    Usage:

    ```python
    process_values = ProcessContigs(**kwargs)
    process_values.process_contigs()
    ```
    """

    def __init__(self, input_file: str = None, genome_database: str = None, output_prefix: str = None,
                 ignore_overlap: bool = True, text_output: bool = False, remove_ccgg: bool = False,
                 min_read_depth: int = 10, max_read_depth: int = 8000, threads: int = 1, verbose: bool = True,
                 min_base_quality: int = 10, min_mapping_quality: int = 10,
                 ATCGmap: bool = False, cg_only: bool = True, ignore_orphans: bool = False,
                 bedgraph_output: bool = False):
        assert isinstance(input_file, str), 'Path to input file not valid'
        assert isinstance(text_output, bool), 'Not valid bool'
        assert isinstance(threads, int), 'Threads must be specified with integer'
        if output_prefix:
            assert isinstance(output_prefix, str)
        self.input_file = input_file
        try:
            self.input_bam = pysam.Samfile(input_file, 'rb', require_index=True)
        except IOError:
            print('Generating Index File')
            pysam.index(input_file)
            self.input_bam = pysam.Samfile(input_file, 'rb', require_index=True)
        self.text_output = text_output
        self.output_prefix = output_prefix
        self.threads = threads
        self.call_methylation_kwargs = dict(input_file=input_file,
                                            genome_database=genome_database,
                                            ignore_overlap=ignore_overlap,
                                            ignore_orphans=ignore_orphans,
                                            remove_ccgg=remove_ccgg,
                                            max_read_depth=max_read_depth,
                                            min_base_quality=min_base_quality,
                                            min_mapping_quality=min_mapping_quality,
                                            cg_only=cg_only)
        self.min_read_depth = min_read_depth
        self.ATCGmap = ATCGmap
        self.bedgraph_output = bedgraph_output
        self.methylation_calling = True
        self.contigs = self.get_contigs
        self.completed_contigs = None
        self.return_queue = None
        self.pool = None
        self.verbose = verbose
        self.output_objects = self.get_output_objects
        self.methylation_stats = {'CG_meth': 0, 'CG_all': 0, 'CH_meth': 0, 'CH_all': 0}

    @property
    def get_contigs(self) -> List[str]:
        """
        get list of contigs in input file, threads set across contigs
        """
        contigs = [contig[0] for contig in self.input_bam.get_index_statistics() if contig[1] > 0]
        if not contigs:
            print('No reads are mapped, exiting methylation calling\nPlease check alignment file\n')
            raise MethylationCallingError
        return contigs

    def process_contigs(self):
        """Launches a processing pool to call methylation values across the input file contigs
        """
        # initialize manager
        manager = multiprocessing.Manager()
        # get return dictionary
        self.return_queue = manager.Queue(maxsize=20)
        self.completed_contigs = manager.list()
        # threads for methylation calling, if one thread use thread for calling and watching
        pool_threads = self.threads - 1 if self.threads != 1 else 1
        # start pool
        self.pool = multiprocessing.Pool(processes=pool_threads)
        # for contig call methylation and return values to dict
        for contig in self.contigs:
            contig_kwargs = dict(self.call_methylation_kwargs)
            contig_kwargs.update(dict(contig=contig, return_queue=self.return_queue))
            self.pool.apply_async(call_contig_methylation,
                                  args=[self.completed_contigs, contig_kwargs],
                                  error_callback=self.methylation_process_error)
        self.pool.close()

    def methylation_process_error(self, error):
        """Raise if exception thrown in methylation calling process"""
        self.methylation_calling = False
        raise MethylationCallingError(error)

    def watch_pool(self):
        """Watch self.return_dict and process return methylation values. Contigs are processed in order so buffer can
        become large if first contig is large, ie. Human Chr1
        """
        contigs_complete = 0
        pbar = None
        if self.verbose:
            pbar = tqdm(total=len(self.contigs), desc='Processing Contigs')
        while self.methylation_calling:
            methylation_lines: list = self.return_queue.get(block=True)
            if self.verbose:
                if len(self.completed_contigs) != contigs_complete:
                    update_number = len(self.completed_contigs) - contigs_complete
                    contigs_complete = len(self.completed_contigs)
                    pbar.update(update_number)
            self.write_output(methylation_lines)
            if len(self.completed_contigs) == len(self.contigs) and self.return_queue.empty():
                self.methylation_calling = False
        if self.verbose:
            pbar.close()
        for out in self.output_objects.values():
            out.close()
        self.print_stats()

    def write_output(self, methylation_lines: List[Tuple[Union[int, float, str]]]):
        """Give a list of methylation call dicts, output formatted line

        Params:

        * *methylation_lines (list)*: list of dict containing methylation call information
        """
        # write wig contig designation
        if methylation_lines:
            for meth_tuple in methylation_lines:
                # unpack methylation data
                meth_line = self.unpack_meth_line(meth_tuple)
                # collect methylation stats
                self.collect_stats(meth_line)
                # write ATCGmap line
                if self.ATCGmap:
                    self.write_line(self.output_objects['ATCGmap'], self.format_atcg(meth_line))
                # if methylation level greater than or equal to min_read_depth output CGmap and wig lines
                if meth_line['all_cytosines'] >= self.min_read_depth:
                    if self.bedgraph_output:
                        self.write_line(self.output_objects['CGmap'], self.format_bedgraph(meth_line))
                    else:
                        self.write_line(self.output_objects['CGmap'], self.format_cgmap(meth_line))

    @staticmethod
    def unpack_meth_line(meth_line: Tuple[Union[int, float, str]]) -> Dict[str, Union[int, float, str]]:
        meth_keys = ['nucleotide', 'meth_cytosines', 'unmeth_cytosines', 'all_cytosines', 'meth_level',
                     'forward_counts', 'reverse_counts', 'pos', 'chrom', 'context', 'subcontext']
        return {key: value for key, value in zip(meth_keys, meth_line)}

    @staticmethod
    def format_atcg(meth_line: Dict[str, Union[int, float, str]]) -> str:
        """ATCGmap line formatting
        """
        ATCGmap_line = f'{meth_line["chrom"]}\t{meth_line["nucleotide"]}\t{meth_line["pos"]}\t{meth_line["context"]}' \
                       f'\t{meth_line["subcontext"]}\t{meth_line["forward_counts"]}\t{meth_line["reverse_counts"]}' \
                       f'\t{meth_line["meth_level"]}\n'
        return ATCGmap_line

    @staticmethod
    def format_cgmap(meth_line: Dict[str, Union[int, float, str]]) -> str:
        """CGmap line formatting
        """
        CGmap_line = f'{meth_line["chrom"]}\t{meth_line["nucleotide"]}\t{meth_line["pos"]}' \
                     f'\t{meth_line["context"]}\t{meth_line["subcontext"]}\t{meth_line["meth_level"]}' \
                     f'\t{meth_line["meth_cytosines"]}\t{meth_line["all_cytosines"]}\n'
        return CGmap_line

    @staticmethod
    def format_bedgraph(meth_line: Dict[str, Union[int, float, str]]) -> str:
        """Bedgraph line formatting"""
        bedgraph_line = f'{meth_line["chrom"]}\t{meth_line["pos"] - 1}\t{meth_line["pos"]}' \
                        f'\t{meth_line["meth_level"] * 100:.3f}\t{meth_line["meth_cytosines"]}' \
                        f'\t{meth_line["unmeth_cytosines"]}\n'
        return bedgraph_line

    @property
    def get_output_objects(self):
        output_objects = {}
        output_path = self.input_file
        output_suffix = '.CGmap' if not self.bedgraph_output else '.bg'
        if self.output_prefix:
            output_path = self.output_prefix
        if self.text_output:
            output_objects['CGmap'] = open(f'{output_path}{output_suffix}', 'w')
            if self.ATCGmap:
                output_objects['ATCGmap'] = open(f'{output_path}.ATCGmap', 'w')
        else:
            output_objects['CGmap'] = io.BufferedWriter(gzip.open(f'{output_path}{output_suffix}.gz', 'wb'))
            if self.ATCGmap:
                output_objects['ATCGmap'] = io.BufferedWriter(gzip.open(f'{output_path}.ATCGmap.gz', 'wb'))
        return output_objects

    def write_line(self, output_object, line: str):
        """ Outputs line, and encodes if necessary

        Params:

        * *output_object (TextIO/GZipIO)*: output object
        * *line (str)*: formatted line to write
        """
        if self.text_output:
            output_object.write(line)
        else:
            output_object.write(line.encode('utf-8'))

    def collect_stats(self, meth_line):
        """Collect global methylation statistics"""
        if meth_line['subcontext'] == 'CG':
            self.methylation_stats['CG_all'] += meth_line['all_cytosines']
            self.methylation_stats['CG_meth'] += meth_line['meth_cytosines']
        else:
            self.methylation_stats['CH_all'] += meth_line['all_cytosines']
            self.methylation_stats['CH_meth'] += meth_line['meth_cytosines']

    def print_stats(self):
        """Print global methylation statistics"""
        try:
            cpg_percentage = (self.methylation_stats["CG_meth"] / self.methylation_stats["CG_all"]) * 100
        except ZeroDivisionError:
            print('Warning! No SAM reads passed quality control metrics')
            raise MethylationCallingError
        try:
            ch_percentage = (self.methylation_stats["CH_meth"] / self.methylation_stats["CH_all"]) * 100
        except ZeroDivisionError:
            ch_percentage = 0.000
        print(f'Methylated CpG Cytosines: {self.methylation_stats["CG_meth"]}')
        print(f'Total Observed CpG Cytosines: {self.methylation_stats["CG_all"]}')
        print(f'Methylated / Total Observed CpG Cytosines: {cpg_percentage:.2f}%')
        print(f'Methylated CH Cytosines: {self.methylation_stats["CH_meth"]}')
        print(f'Total Observed CH Cytosines: {self.methylation_stats["CH_all"]}')
        print(f'Methylated / Total Observed CH Cytosines: {ch_percentage:.2f}%')



import os
import subprocess
from typing import List
from BSBolt.Utils.UtilityFunctions import get_external_paths


class BisulfiteAlignmentAndProcessing:
    """ Read alignment as processing. Handles reads from BWA-MEM2 and outputs to BAM file. Also aggregates alignment
    stats.

    Params:

    * *alignment_commands (list)*: bwa alignment commands
    * *output (str)*: output prefix

    Attributes:

    * self.mapping_statistics (dict)*: alignment run statistics
    """

    def __init__(self, alignment_commands: List[str], output=None):
        self.alignment_commands = alignment_commands
        self.output = output
        self.mapping_statistics = dict(TotalReads=0, TotalAlignments=0, BSAmbiguous=0, C_C2T=0, C_G2A=0,
                                       W_C2T=0, W_G2A=0, Unaligned=0)

    def align_reads(self):
        """ Launch bwa alignment. Pipe output to BAM file
        """
        _, _, stream_bam = get_external_paths()
        if '/' in self.output:
            assert os.path.exists('/'.join(self.output.split('/')[0:-1])), f"output path {self.output} not valid"
        alignment_run = subprocess.Popen(self.alignment_commands,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE,
                                         universal_newlines=True)
        bam_compression = subprocess.Popen([stream_bam, '-o', f'{self.output}.bam'],
                                           stdin=alignment_run.stdout)
        # watch alignment progress, output stderr and collect alignment stats
        while True:
            if bam_compression.returncode:
                break
            elif alignment_run.returncode:
                break
            elif alignment_run.poll() is not None and bam_compression.poll() is not None:
                alignment_run.stdout.close()
                alignment_run.stderr.close()
                break
            else:
                alignment_info = alignment_run.stderr.readline().strip()
                if alignment_info:
                    if alignment_info[0:7] == 'BSStat ':
                        category, count = alignment_info.replace('BSStat ', '').split(': ')
                        self.mapping_statistics[category] += int(count)
                    else:
                        print(alignment_info)



import subprocess
from typing import Dict, Union
from BSBolt.Utils.UtilityFunctions import retrieve_iupac


class StreamSim:

    def __init__(self, paired_end=False, sim_command=None):
        self.paired_end = paired_end
        self.sim_command = sim_command
        self.contig_variants = {}
        self.variant_contig = None

    def __iter__(self):
        sim = subprocess.Popen(self.sim_command,
                               stdout=subprocess.PIPE,
                               universal_newlines=True)
        variant_output = False
        sim_output = iter(sim.stdout.readline, '')
        read_pair = {1: None, 2: None}
        paired_count = 0
        while True:
            try:
                formatted_line = next(sim_output).strip()
            except StopIteration:
                break
            else:
                if formatted_line == 'Contig Variant Start':
                    variant_output = True
                elif variant_output:
                    variant_output = self.collect_variant_info(formatted_line)
                    if not variant_output and self.contig_variants:
                        yield self.variant_contig, self.contig_variants
                        self.contig_variants = {}
                else:
                    read_info = self.process_read_name(formatted_line)
                    seq = next(sim_output).strip()
                    comment = next(sim_output).strip()
                    qual = self.modify_qual(next(sim_output).strip())
                    read_info.update(dict(comment=comment, seq=seq, qual=qual))
                    read_pair[read_info['pair']] = read_info
                    paired_count += 1
                    if paired_count == 2:
                        assert read_pair[1]['read_id'] == read_pair[2]['read_id']
                        yield False, read_pair
                        read_pair = {1: None, 2: None}
                        paired_count = 0

    @staticmethod
    def modify_qual(quality: str) -> str:
        # modify start position to ensure proper qual handling downstream
        quality_split = list(quality)
        qual_start = int(ord(quality_split[0]) - 33)
        quality_split[0] = chr(qual_start + 32)
        return ''.join(quality_split)

    def collect_variant_info(self, formatted_line):
        if formatted_line == 'Contig Variant End':
            return False
        variant_info = self.process_variant_line(formatted_line)
        assert variant_info['pos'] not in self.contig_variants
        self.contig_variants[variant_info['pos']] = variant_info
        self.variant_contig = variant_info['chrom']
        return True

    @staticmethod
    def process_variant_line(formatted_line):
        chrom, pos, reference, alt, heterozygous = formatted_line.split('\t')
        pos = int(pos)
        indel = 0
        iupac = None
        if reference == '-':
            indel = 1
        elif alt == '-':
            indel = -1
        else:
            iupac = retrieve_iupac(alt)
        return dict(chrom=chrom, pos=pos, reference=reference, alt=alt, heterozygous=heterozygous,
                    indel=indel, iupac=iupac)

    @staticmethod
    def process_read_name(formatted_line: str) -> Dict[str, Union[str, int]]:
        read_info = formatted_line.split(':')
        chrom, start, end, insert_size, read_id, cigar, pair, c_base_info, g_base_info = read_info
        return dict(chrom=chrom.replace('@', ''), start=int(start), end=int(end),
                    insert_size=insert_size, read_id=read_id, cigar=cigar, pair=int(pair),
                    c_base_info=c_base_info, g_base_info=g_base_info)



import pickle
import re
from typing import Any, Dict, List, Tuple, Union
import pysam
import numpy as np


class CallMethylationVector:
    """
     Experimental:

     return vector of methylated sites
    """

    def __init__(self, input_file: str = None, genome_database: str = None,
                 contig: str = None, min_base_quality: int = 0, min_mapping_quality: int = 0, return_queue=None,
                 cg_only: bool = False, start=None, end=None, filter_duplicates=True):
        self.input_file = str(input_file)
        self.input_bam = pysam.AlignmentFile(self.input_file, 'rb',
                                             require_index=True)
        self.genome_database = str(genome_database)
        if self.genome_database[-1] != '/':
            self.genome_database = f'{self.genome_database}/'
        self.contig = contig
        self.start = start
        self.mate_flags = self.get_mate_flags
        self.filter_duplicates = filter_duplicates
        self.end = end
        self.min_base_quality = min_base_quality
        self.min_mapping_quality = min_mapping_quality
        self.chunk_size = 10000
        self.cg_only = cg_only
        self.return_queue = return_queue
        self.counting_dict = {}

    @property
    def get_mate_flags(self) -> Dict[int, int]:
        # mate flag check for read pairing
        mate_flags = {67: 131, 323: 387, 115: 179, 371: 435,
                      131: 67, 387: 323, 179: 115, 435: 371}
        return mate_flags

    def call_methylation(self):
        try:
            chrom_seq = self.get_reference_sequence(f'{self.genome_database}{self.contig}.pkl')
        except FileNotFoundError:
            self.return_queue.put([])
            print(f'{self.contig} not found in BSBolt DB, Methylation Calls for {self.contig} skipped. Methylation '
                  f'values should be called using the same DB used for alignment.')
            self.return_queue.put([])
        else:
            self.call_contig(chrom_seq)

    def call_contig(self, chrom_seq: str):
        """Iterates through bam reads call methylation along vectors. When a read overlaps a site where methylation is
         called the site with a higher quality is taken. If overlapping sites with the same quality are observed
         the first observed site is reported. If an overlapping site is reported as a mismatch only the site with
         a methylation call is reported (this should be extremely rare but is observed in the test cases)
        """
        # set search pattern for only CG sites or all Cs
        search_pattern = ('C', 'G')
        if self.cg_only:
            search_pattern = ('CG', 'CG')
        # iterate through pileup
        contig_chunk = []
        methylation_vectors = {}
        for aligned_read in self.input_bam.fetch(contig=self.contig, start=self.start, end=self.end,
                                                 multiple_iterators=True):
            if aligned_read.is_unmapped:
                continue
            if aligned_read.is_duplicate and self.filter_duplicates:
                continue
            if aligned_read.mapping_quality < self.min_mapping_quality:
                continue
            # get sequence around pileup site
            reference_start = aligned_read.reference_start
            reference_seq = chrom_seq[reference_start - 1: aligned_read.reference_end + 2].upper()
            c_search_pattern, reference_nuc, strand, offset = search_pattern[0], 'C', 'watson', 0
            if aligned_read.is_reverse:
                c_search_pattern, reference_nuc, strand = search_pattern[1], 'G', 'crick'
                if self.cg_only:
                    offset = 1
            c_search = [match.start() + offset + reference_start - 1 for match in
                        re.finditer(c_search_pattern, reference_seq)]
            # if reference sequence not found proceed to next sequence
            if not c_search:
                continue
            methylation_calls = self.call_vector(aligned_read, set(c_search), reference_nuc)
            if not methylation_calls[0]:
                continue
            processed_vector = self.process_methylation_vector(aligned_read, methylation_calls,
                                                               strand, methylation_vectors)
            if processed_vector:
                contig_chunk.append(processed_vector)
            if len(contig_chunk) == self.chunk_size:
                self.return_queue.put((self.contig, contig_chunk))
                contig_chunk = []
        # process reads that didn't have a pair with a observed methylation site
        for call in methylation_vectors.values():
            contig_chunk.append((call['read_name'], call['calls'][1][0], call['calls'][1][-1],
                                 np.asarray(call['calls'][0]), np.asarray(call['calls'][1]), call['flag'],
                                 self.mate_flags[call['flag']], call['strand']))
        if contig_chunk:
            self.return_queue.put((self.contig, contig_chunk))

    def call_vector(self, aligned_read: pysam.AlignedRead,
                    positions: set,
                    reference_nuc: str) -> List:
        methylation_calls = [[], [], []]
        reference_consumers = {0, 2, 3, 7, 8}
        query_consumers = {0, 1, 4, 7, 8}
        # set relative to genomic position so add reference start and one since capturing the first base
        reference_pos = aligned_read.reference_start
        query_sequence = aligned_read.query_sequence
        query_qualities = aligned_read.query_qualities
        query_position = 0
        for cigar_type, cigar_count in aligned_read.cigartuples:
            if cigar_type in reference_consumers and cigar_type in query_consumers:
                for _ in range(cigar_count):
                    if reference_pos in positions:
                        pos_qual = query_qualities[query_position]
                        if pos_qual > self.min_base_quality:
                            query_base = query_sequence[query_position]
                            call_made, methylation_value = self.get_methylation_call(reference_nuc, query_base)
                            if call_made:
                                methylation_calls[0].append(methylation_value)
                                methylation_calls[1].append(reference_pos)
                                methylation_calls[2].append(pos_qual)
                    reference_pos += 1
                    query_position += 1
            elif cigar_type in reference_consumers and cigar_type not in query_consumers:
                reference_pos += cigar_count
            elif cigar_type in query_consumers and cigar_type not in reference_consumers:
                query_position += cigar_count
        return methylation_calls

    def process_methylation_vector(self,
                                   aligned_read: pysam.AlignedRead,
                                   methylation_calls: List,
                                   strand: str, methylation_vectors: Dict[str, Any]) -> Union[None, Tuple]:
        if aligned_read.is_proper_pair:
            mate_flag = self.mate_flags[aligned_read.flag]
            mate_pair_label = f'{aligned_read.query_name}_{mate_flag}_{aligned_read.next_reference_start}'
            if mate_pair_label in methylation_vectors:
                paired_calls = methylation_vectors.pop(mate_pair_label)['calls']
                paired_calls[0].extend(methylation_calls[0])
                paired_calls[1].extend(methylation_calls[1])
                paired_calls[2].extend(methylation_calls[2])
                paired_calls = self.clean_overlap(paired_calls)
                return (aligned_read.query_name, paired_calls[1][0], paired_calls[1][-1],
                        np.array(paired_calls[0]), np.array(paired_calls[1]),
                        aligned_read.flag, mate_flag, strand)
            else:
                vector_label = f'{aligned_read.query_name}_{aligned_read.flag}_{aligned_read.reference_start}'
                methylation_vectors[vector_label] = {'calls': methylation_calls,
                                                     'read_name': aligned_read.query_name,
                                                     'flag': aligned_read.flag, 'strand': strand}
                return None
        else:
            return (aligned_read.query_name, methylation_calls[1][0], methylation_calls[1][-1],
                    np.array(methylation_calls[0]), np.array(methylation_calls[1]), aligned_read.flag, None, strand)

    @staticmethod
    def clean_overlap(methylation_calls: List) -> List:
        cleaned_calls = {}
        for meth_call, pos, qual in zip(methylation_calls[0], methylation_calls[1], methylation_calls[2]):
            if pos not in cleaned_calls:
                cleaned_calls[pos] = (meth_call, qual)
            else:
                if qual > cleaned_calls[pos][1]:
                    cleaned_calls[pos] = (meth_call, qual)
        return [[call[0] for call in cleaned_calls.values()], list(cleaned_calls.keys())]

    @staticmethod
    def get_methylation_call(reference_nucleotide: str, base_call: str) -> Tuple[bool, Union[None, int]]:
        """
        Methylation for a C relative to the sense strand of the reference can only be called using watson reads,
        and G with crick reads
        Arguments
            nucleotide (str): reference nucleotide
            base_call (collections.Counter): watson nucleotides are Uppercase and crick nucleotides lowercase
        Returns:
             methylation call dictionary
        """
        # call cytosine with watson
        if reference_nucleotide == 'C':
            if base_call == 'C':
                return True, 1
            elif base_call == 'T':
                return True, 0
        elif reference_nucleotide == 'G':
            if base_call == 'G':
                return True, 1
            elif base_call == 'A':
                return True, 0
        return False, None

    @staticmethod
    def get_reference_sequence(path: str) -> str:
        """load serialized reference file from path
        """
        with open(path, 'rb') as genome_file:
            return pickle.load(genome_file)



#! /usr/bin/env python3

from BSBolt.Impute.kNN_Impute import ImputeMissingValues
from BSBolt.Impute.Validation.MaskValues import MaskImputationValues


class ImputationAssessment:
    """

           :param methylation_array:
           :param masking_proportion:
           :param masking_sites:
           :param verbose:
           :param k:
           :param imputation_window_size:
           :param threads:
           :param methylation_site_order:
           :param methylation_samples:
           :param batch_size:
           :param randomize_batch:
           """

    def __init__(self, methylation_array=None, masking_proportion=0.1, masking_sites=None, verbose=False,
                 k=5, imputation_window_size=3000000, threads=4, methylation_site_order=None, methylation_samples=None,
                 batch_size=None, randomize_batch=False):
        self.methylation_array = methylation_array
        masking_kwargs = dict(methylation_array=methylation_array, masking_proportion=masking_proportion,
                              masking_sites=masking_sites, verbose=verbose)
        self.masking = MaskImputationValues(**masking_kwargs)
        self.masking_sites = masking_sites
        self.imputation_kwargs = dict(k=k, imputation_window_size=imputation_window_size, threads=threads,
                                      verbose=verbose, meth_site_order=methylation_site_order,
                                      sample_ids=methylation_samples, batch_size=batch_size,
                                      randomize_batch=randomize_batch)
        self.stats = {}
        self.imputed_array = None

    def evaluate_imputation(self):
        if self.masking_sites:
            self.masking.mask_known_sites()
        else:
            self.masking.mask_random_sites()
        self.imputation_kwargs.update(dict(meth_matrix=self.masking.methylation_array))
        self.impute_masked_values()
        self.get_imputation_statistics()

    def impute_masked_values(self):
        masked_imputation = ImputeMissingValues(**self.imputation_kwargs)
        masked_imputation.impute_values()
        self.imputed_array = masked_imputation.meth_matrix

    def get_imputation_statistics(self):
        for site, site_value in self.masking.masking_sites.items():
            row_index, column_index = (int(x) for x in site.split('_'))
            imputed_value = self.imputed_array[row_index, column_index]
            absolute_error = abs(site_value - imputed_value)
            self.stats[site] = dict(site_value=site_value, imputed_value=imputed_value, absolute_error=absolute_error)



from typing import List, Tuple
from BSBolt.Utils.UtilityFunctions import reverse_complement, retrieve_iupac


class ProcessCutSites:
    """ Process cut format information and returns a dictionary with restriction sequence, reverse complement of
    restriction sequence and offset.

    Params:

    * *cut_format (str)*: restriction enzyme recognition sequence, [C-CGG]
    * *cut_descriptor (str)*: character used to designate cut site, [-]

    """

    def __init__(self, cut_format: str = None, cut_descriptor: str = '-'):
        self.cut_format: list = cut_format.upper().replace(' ', '').split(',')
        self.cut_descriptor = cut_descriptor
        self.restriction_site_dict = {}
        self.process_cut_sites()

    def process_cut_sites(self):
        """Process forward and reverse strand restriction sites
        """
        for site in self.cut_format:
            # get offset for restriction sequence
            forward_offset, reverse_offset = self.get_site_offsets(site)
            # remove cut descriptor from recognition sequence
            nucleotide_site = site.replace(self.cut_descriptor, '')
            # if iupac symbol in restriction site get all possible recognition sequences
            forward_sites, reverse_sites = self.get_recognition_site_sequences(nucleotide_site)
            # iterate through site and population self.restriction_site_dict
            for f_site, r_site in zip(forward_sites, reverse_sites):
                self.restriction_site_dict[f_site] = forward_offset
                # if the f_site is equivalent to the r_site don't add reverse_site as the offset will be incorrect
                if f_site != r_site:
                    self.restriction_site_dict[r_site] = reverse_offset

    def get_site_offsets(self, site: str) -> Tuple[int, int]:
        """Given a restriction site return offset, or proper position based on where cut occurs in sequence

        Params:

        * *site (str)*: recognition site sequence

        Returns:

        * *forward_offest (int/bool)*: if cut descriptor int, else False
        * *reverse_offset (int/bool)*: if cut descriptor int, else False
        """
        try:
            # retrieve index of cut descriptor where present
            forward_offset = site.index(self.cut_descriptor)
        except ValueError:
            # if descriptor not present set to 0
            return 0, 0
        else:
            # get location of cut_descriptor in reverse complement
            reverse_offset = reverse_complement(site).index(self.cut_descriptor)
            return forward_offset, reverse_offset

    def get_recognition_site_sequences(self, recognition_site: str) -> Tuple[List[str], List[str]]:
        """
        Params:

        * *recognition_site (str)*: sequence of site

        Returns:

        * *forward_recognition_sites (list)*: list of forward recognition sites
        * *reverse_recognition_sites (list)*: list of reverse recognition sites
        """
        forward_recognition_sites = ['']
        # iterate through nucleotides in recognition site and replace cut_descriptor
        for nucelotide in recognition_site.replace(self.cut_descriptor, ''):
            temp_sites = []
            # retrieve_iupac returns value of iupac identifiers or the original nucleotide
            for iupac_nucleotide in retrieve_iupac(nucelotide):
                for site in forward_recognition_sites:
                    temp_sites.append(f'{site}{iupac_nucleotide}')
            forward_recognition_sites = temp_sites
        # get reverse complement for all forward recognition sites
        reverse_recognition_sites = [reverse_complement(site) for site in forward_recognition_sites]
        return forward_recognition_sites, reverse_recognition_sites



from collections import Counter
import multiprocessing as mp
from typing import List
from BSBolt.Utils.CGmapIterator import OpenCGmap
from BSBolt.Utils.UtilityFunctions import propagate_error
from tqdm import tqdm


class CGmapSiteCollector:
    """Iterate through provided CGmap files counting the number of times a specific site is observed between all the
       cgmap files.
       Keyword Arguments
           cgmap_files (list): list of cgmap file paths
           min_site_coverage (int): min site coverage for a valid observation
           cg_only (bool): only consider cg context sites
           batch_size (int): chunk size for methylation observations to return
           verbose (bool): tqdm output
           threads (int): number of threads available
           site_proportion_threshold (int): number of times a site must be observed to be included in matrix
    """

    def __init__(self, cgmap_files: List[str] = None,
                 min_site_coverage: int = 1,
                 cg_only: bool = False,
                 batch_size: int = 10000,
                 verbose: bool = False,
                 threads: int = 8,
                 site_proportion_threshold: int = 1):
        self.cgmap_files = cgmap_files
        self.min_site_coverage = min_site_coverage
        self.cg_only = cg_only
        self.batch_size = batch_size
        self.verbose = verbose
        self.threads = threads
        self.site_proportion_threshold = site_proportion_threshold
        self.observed_site_count = Counter()

    def collect_consensus_sites(self) -> List[str]:
        # initialize collection_pool and manager queue
        collection_pool = mp.Pool(processes=self.threads)
        manager = mp.Manager()
        collection_queue = manager.Queue(maxsize=10)
        completed_samples = manager.list()
        for cgmap in self.cgmap_files:
            collection_pool.apply_async(self.collect_methylation_sites,
                                        kwds=dict(cgmap_path=cgmap, return_queue=collection_queue,
                                                  completed_samples=completed_samples),
                                        error_callback=propagate_error)
        collection_pool.close()
        # if verbose initialize tqdm progress bar
        update_count = 0
        pbar = tqdm(total=len(self.cgmap_files), desc='Collecting Methylation Sites') if self.verbose else None
        while True:
            if len(completed_samples) == len(self.cgmap_files) and collection_queue.empty():
                break
            sites = collection_queue.get(block=True)
            self.observed_site_count.update(sites)
            if pbar:
                update_number = len(completed_samples) - update_count
                pbar.update(update_number)
                update_count = len(completed_samples)
        if pbar:
            update_number = len(completed_samples) - update_count
            pbar.update(update_number)
            pbar.close()
        return [site for site, count in self.observed_site_count.items() if count >= self.site_proportion_threshold]

    def collect_methylation_sites(self, cgmap_path: str = None, return_queue: mp.Queue = None,
                                  completed_samples: List = None):
        site_count, methylation_sites = 0, []
        for line_info in OpenCGmap(cgmap_path):
            try:
                line_info[7]
            except IndexError:
                break
            else:
                if self.check_cg(line_info[4]):
                    if int(line_info[7]) >= self.min_site_coverage:
                        methylation_sites.append(f'{line_info[0]}:{line_info[2]}')
                        site_count += 1
            if site_count == self.batch_size:
                return_queue.put(tuple(methylation_sites), block=True)
                site_count, methylation_sites = 0, []
        if methylation_sites:
            return_queue.put(tuple(methylation_sites), block=True)
        completed_samples.append('done')

    def check_cg(self, nucleotide_context: str) -> bool:
        if self.cg_only:
            if nucleotide_context == 'CG':
                return True
            return False
        return True
